@@ -32,7 +32,7 @@
 #ifndef PYGETDNS_H
 #define PYGETDNS_H
 
-#define PYGETDNS_VERSION "0.3.1"
+#define PYGETDNS_VERSION "0.5.0"
 #define GETDNS_DOCSTRING "getdns bindings for Python (see http://www.getdnsapi.net)"
 
 #define GETDNS_STR_IPV4 "IPv4"
@@ -59,6 +59,11 @@ typedef struct  {
     PyObject *canonical_name;
     PyObject *replies_full;
     PyObject *validation_chain;
+#if GETDNS_NUMERIC_VERSION < 0x00090000
+    PyObject *call_debugging;
+#else 
+    PyObject *call_reporting;
+#endif
 } getdns_ResultObject;
 
 
@@ -73,8 +78,11 @@ typedef struct {
     PyObject_HEAD
     PyObject *py_context;       /* Python capsule containing getdns_context */
     uint64_t  timeout;          /* timeout attribute (milliseconds) */
+    uint64_t  idle_timeout;     /* TCP timeout attribute (milliseconds) */
     getdns_resolution_t resolution_type; /* stub or recursive? */
+#if 0
     getdns_transport_t dns_transport;    /* udp/tcp/etc */
+#endif    
     uint16_t limit_outstanding_queries;
     getdns_redirects_t follow_redirects;
     getdns_append_name_t append_name;
@@ -88,11 +96,21 @@ typedef struct {
     getdns_list *dns_root_servers;
     getdns_list *dnssec_trust_anchors;
     getdns_list *upstream_recursive_servers;
+    getdns_transport_list_t *dns_transport_list;
     struct event_base *event_base;
     char *implementation_string;
     char *version_string;
+    uint16_t tls_authentication;
+#if GETDNS_NUMERIC_VERSION > 0x00050000
+    uint16_t tls_query_padding_blocksize;
+    uint8_t edns_client_subnet_private;
+#endif
 } getdns_ContextObject;
 
+struct setter_table  {          /* we're now using bsearch to find */
+    char *name;                 /* setters for attribute names - somewhat */
+    int (*setter)(getdns_context *, PyObject *);            /* more efficient but much more maintainable */
+};
 
 extern PyTypeObject getdns_ResultType;
 void result_dealloc(getdns_ResultObject *self);
@@ -107,6 +125,11 @@ char *get_canonical_name(struct getdns_dict *result_dict);
 PyObject *get_just_address_answers(struct getdns_dict *result_dict);
 PyObject *get_replies_tree(struct getdns_dict *result_dict);
 PyObject *get_validation_chain(struct getdns_dict *result_dict);
+#if GETDNS_NUMERIC_VERSION < 0x00090000
+PyObject *get_call_debugging(struct getdns_dict *result_dict);
+#else
+PyObject *get_call_reporting(struct getdns_dict *result_dict);
+#endif
 
 int context_init(getdns_ContextObject *self, PyObject *args, PyObject *keywds);
 PyObject *context_getattro(PyObject *self, PyObject *nameobj);
@@ -127,6 +150,9 @@ int context_set_dns_root_servers(getdns_context *context, PyObject *py_value);
 int context_set_dnssec_trust_anchors(getdns_context *context, PyObject *py_value);
 int context_set_upstream_recursive_servers(getdns_context *context, PyObject *py_value);
 int context_set_edns_do_bit(getdns_context *context, PyObject *py_value);
+int context_set_dns_transport_list(getdns_context *context, PyObject *py_value);
+int context_set_tls_query_padding_blocksize(getdns_context *context, PyObject *py_value);
+int context_set_edns_client_subnet_private(getdns_context *context, PyObject *py_value);
 
 PyObject *context_str(PyObject *self);
 
@@ -138,6 +164,8 @@ PyObject *context_service(getdns_ContextObject *self, PyObject *args, PyObject *
 PyObject *context_run(getdns_ContextObject *self, PyObject *args, PyObject *keywds);
 PyObject *context_cancel_callback(getdns_ContextObject *self, PyObject *args, PyObject *keywds);
 
+PyObject *context_get_attributes(getdns_ContextObject *self, PyObject *unused);
+
 void context_dealloc(getdns_ContextObject *self);
 PyObject *get_callback(char *py_main, char *callback);
 void callback_shim(struct getdns_context *context, getdns_callback_type_t type,
@@ -157,5 +185,6 @@ PyObject *decode_getdns_replies_tree_response(struct getdns_dict *response);
 PyObject *getFullResponse(struct getdns_dict *dict);
 getdns_dict *getdnsify_addressdict(PyObject *pydict);
 
+PyObject *convertToDict(struct getdns_dict* dict);
 
 #endif /* PYGETDNS_H */
