@@ -1,4 +1,4 @@
- /*
+/*
  * \ file pygetdns_util.c
  * @brief utility functions to support pygetdns bindings
  */
@@ -37,7 +37,6 @@
 #include <string.h>
 #include <getdns/getdns.h>
 #include <getdns/getdns_extra.h>
-#include <ldns/ldns.h>
 #include "pygetdns.h"
 
 
@@ -123,6 +122,37 @@ get_validation_chain(struct getdns_dict *result_dict)
 }
 
 
+#if GETDNS_NUMERIC_VERSION < 0x00090000
+PyObject *
+get_call_debugging(struct getdns_dict *result_dict)
+#else
+PyObject *
+get_call_reporting(struct getdns_dict *result_dict)
+#endif
+{
+#if GETDNS_NUMERIC_VERSION < 0x00090000
+    struct getdns_list *call_debugging;
+#else
+    struct getdns_list *call_reporting;
+#endif
+    getdns_return_t ret;
+
+
+#if GETDNS_NUMERIC_VERSION < 0x00090000
+    if ((ret = getdns_dict_get_list(result_dict, "call_debugging", &call_debugging)) !=
+#else
+    if ((ret = getdns_dict_get_list(result_dict, "call_reporting", &call_reporting)) !=
+#endif
+        GETDNS_RETURN_GOOD)
+        Py_RETURN_NONE;
+    else
+#if GETDNS_NUMERIC_VERSION < 0x00090000
+        return glist_to_plist(call_debugging);
+#else
+        return glist_to_plist(call_reporting);
+#endif
+}
+
 struct getdns_dict *
 extensions_to_getdnsdict(PyDictObject *pydict)
 {
@@ -148,32 +178,58 @@ extensions_to_getdnsdict(PyDictObject *pydict)
         char *tmp_key;
         int  tmp_int;
 
+#if PY_MAJOR_VERSION >= 3
+        tmp_key = PyBytes_AsString(PyUnicode_AsEncodedString(PyObject_Str(key), "ascii", NULL));
+#else
         tmp_key = PyString_AsString(PyObject_Str(key));
+#endif
         if ( (!strncmp(tmp_key, "dnssec_return_status", strlen("dnssec_return_status")))  ||
              (!strncmp(tmp_key, "dnssec_return_only_secure", strlen("dnssec_return_only_secure")))  ||
              (!strncmp(tmp_key, "dnssec_return_validation_chain", strlen("dnssec_return_validation_chain")))  ||
              (!strncmp(tmp_key, "return_both_v4_and_v6", strlen("return_both_v4_and_v6")))  ||
-             (!strncmp(tmp_key, "dnssec_return_supporting_responses", strlen("dnssec_return_supporting_responses")))  ||
              (!strncmp(tmp_key, "return_api_information", strlen("return_api_information")))  ||
+#if GETDNS_NUMERIC_VERSION < 0x00090000
              (!strncmp(tmp_key, "return_call_debugging", strlen("return_call_debugging")))  ||
+#else
+             (!strncmp(tmp_key, "return_call_reporting", strlen("return_call_reporting")))  ||
+#endif
              (!strncmp(tmp_key, "add_warning_for_bad_dns", strlen("add_warning_for_bad_dns"))) )  {
+#if PY_MAJOR_VERSION >= 3
+            if (!PyLong_Check(value))  {
+#else
             if (!PyInt_Check(value))  {
+#endif
                 PyErr_SetString(getdns_error, GETDNS_RETURN_EXTENSION_MISFORMAT_TEXT);
                 return NULL;
             }
+#if PY_MAJOR_VERSION >= 3
+            if ( !((PyLong_AsLong(value) == GETDNS_EXTENSION_TRUE) || (PyLong_AsLong(value) == GETDNS_EXTENSION_FALSE)) )  {
+#else
             if ( !((PyInt_AsLong(value) == GETDNS_EXTENSION_TRUE) || (PyInt_AsLong(value) == GETDNS_EXTENSION_FALSE)) )  {
+#endif
                 PyErr_SetString(getdns_error, GETDNS_RETURN_EXTENSION_MISFORMAT_TEXT);
                 return NULL;
             }
+#if PY_MAJOR_VERSION >= 3
+            tmp_int = (int)PyLong_AsLong(value);
+#else
             tmp_int = (int)PyInt_AsLong(value);
-
+#endif
             (void)getdns_dict_set_int(newdict, tmp_key, tmp_int);
         } else if (!strncmp(tmp_key, "specify_class", strlen("specify_class")))  { /* takes integer */
+#if PY_MAJOR_VERSION >= 3
+            if (!PyLong_Check(value))  {
+#else
             if (!PyInt_Check(value))  {
+#endif
                 PyErr_SetString(getdns_error, GETDNS_RETURN_EXTENSION_MISFORMAT_TEXT);
                 return NULL;
             }
+#if PY_MAJOR_VERSION >= 3
+            tmp_int = (int)PyLong_AsLong(value);
+#else
             tmp_int = (int)PyInt_AsLong(value);
+#endif
             (void)getdns_dict_set_int(newdict, tmp_key, tmp_int);
 
 /*
@@ -198,16 +254,28 @@ extensions_to_getdnsdict(PyDictObject *pydict)
             }
             out_optdict = getdns_dict_create();
             while (PyDict_Next((PyObject *)in_optdict, &opt_pos, &opt_key, &opt_value))  {
+#if PY_MAJOR_VERSION >= 3
+                tmp_opt_key = PyBytes_AsString(PyUnicode_AsEncodedString(opt_key, "ascii", NULL));
+#else
                 tmp_opt_key = PyString_AsString(opt_key);
+#endif
                 if ( (!strncmp(tmp_opt_key, "maximum_udp_payload_size", strlen("maximum_udp_payload_size")))  ||
                      (!strncmp(tmp_opt_key, "extended_rcode", strlen("extended_rcode"))) ||
                      (!strncmp(tmp_opt_key, "version", strlen("version"))) ||
                      (!strncmp(tmp_opt_key, "do_bit", strlen("do_bit"))) )  {
+#if PY_MAJOR_VERSION >= 3
+                    if (!PyLong_Check(opt_value))  {
+#else
                     if (!PyInt_Check(opt_value))  {
+#endif
                         PyErr_SetString(getdns_error, GETDNS_RETURN_EXTENSION_MISFORMAT_TEXT);
                         return NULL;
                     }
+#if PY_MAJOR_VERSION >= 3
+                    optint = (int)PyLong_AsLong(opt_value);
+#else
                     optint = (int)PyInt_AsLong(opt_value);
+#endif
                     (void)getdns_dict_set_int(out_optdict, tmp_opt_key, optint);
                 }  else if (!strncmp(tmp_opt_key, "options", strlen("options")))  { /* options */
 /*
@@ -238,17 +306,36 @@ extensions_to_getdnsdict(PyDictObject *pydict)
                         }
                         /* optionitem should be a dict with keys option_code and option_data */
                         while (PyDict_Next(optionitem, &optiondictpos, &optiondictkey, &optiondictvalue))  {
+#if PY_MAJOR_VERSION >= 3
+                            tmpoptionlistkey = PyBytes_AsString(PyUnicode_AsEncodedString(PyObject_Str(optiondictkey),
+                                                                                          "ascii", NULL)); /* XXX */
+#else
                             tmpoptionlistkey = PyString_AsString(PyObject_Str(optiondictkey));
+#endif
                             if  (!strncmp(tmpoptionlistkey, "option_code", strlen("option_code")))  {
+#if PY_MAJOR_VERSION >= 3
+                                if (!PyLong_Check(optiondictvalue))  {
+#else
                                 if (!PyInt_Check(optiondictvalue))  {
+#endif
                                     PyErr_SetString(getdns_error, GETDNS_RETURN_EXTENSION_MISFORMAT_TEXT);
                                     return NULL;
                                 }
+#if PY_MAJOR_VERSION >= 3
+                                getdns_dict_set_int(tmpoptions_list_dict, "option_code", (uint32_t)PyLong_AsLong(optiondictvalue));
+#else
                                 getdns_dict_set_int(tmpoptions_list_dict, "option_code", (uint32_t)PyInt_AsLong(optiondictvalue));
+#endif
                             }  else if (!strncmp(tmpoptionlistkey, "option_data", strlen("option_data")))  {
                                 option_data = (struct getdns_bindata *)malloc(sizeof(struct getdns_bindata));
                                 option_data->size = PyObject_Length(optiondictvalue);
-                                option_data->data = (uint8_t *)PyString_AsString(PyObject_Bytes(optiondictvalue)); /* This is almost certainly wrong */
+#if PY_MAJOR_VERSION >= 3
+/* This is almost certainly wrong */
+                                option_data->data = (uint8_t *)PyBytes_AsString(PyObject_Bytes(optiondictvalue));
+#else
+/* ditto */
+                                option_data->data = (uint8_t *)PyString_AsString(PyObject_Bytes(optiondictvalue)); 
+#endif
                                 getdns_dict_set_bindata(tmpoptions_list_dict, "option_data", option_data);
                             } else  {
                                 PyErr_SetString(getdns_error, GETDNS_RETURN_EXTENSION_MISFORMAT_TEXT);
@@ -276,7 +363,7 @@ extensions_to_getdnsdict(PyDictObject *pydict)
 
 
 getdns_dict *
-getdnsify_addressdict(PyObject *pydict)
+    getdnsify_addressdict(PyObject *pydict)
 {
     getdns_dict *addr_dict;
     getdns_bindata addr_data;
@@ -284,26 +371,53 @@ getdnsify_addressdict(PyObject *pydict)
     PyObject *str;
     unsigned char buf[sizeof(struct in6_addr)];
     int domain;
-
+    getdns_bindata tls_auth_name;
+    getdns_bindata scope_id;
+    getdns_bindata tsig_name;
+    getdns_bindata tsig_alg;
+    getdns_bindata tsig_secret;
+    uint32_t tls_port, port;
+    getdns_return_t ret;
+    getdns_list *tls_pubkey_pinset;
 
     if (!PyDict_Check(pydict))  {
         PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
         return NULL;
     }
-    if (PyDict_Size(pydict) != 2)  {
-        PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
-        return NULL;
-    }
+
+    /* XXX rewrite this so it's more general */
+
+    /* dict members supported:
+     *  address_data
+     *  address_type
+     *  tls_auth_name
+     *  scope_id
+     *  port
+     *  tls_port
+     *  tsig_name
+     *  tsig_secret
+     *  tsig_algorithm
+     *  tls_pubkey_pinset
+     */
+
     addr_dict = getdns_dict_create();
     if ((str = PyDict_GetItemString(pydict, "address_type")) == NULL)  {
         PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
         return NULL;
     }
+#if PY_MAJOR_VERSION >= 3
+    if (!PyUnicode_Check(str))  {
+#else
     if (!PyString_Check(str))  {
+#endif
         PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
         return NULL;
     }
+#if PY_MAJOR_VERSION >= 3
+    addr_type.data = (uint8_t *)strdup(PyBytes_AsString(PyUnicode_AsEncodedString(str, "ascii", NULL)));
+#else
     addr_type.data = (uint8_t *)strdup(PyString_AsString(str));
+#endif
     addr_type.size = strlen((char *)addr_type.data);
     if (strlen((char *)addr_type.data) != 4)  {
         PyErr_SetString(getdns_error, GETDNS_RETURN_WRONG_TYPE_REQUESTED_TEXT);
@@ -323,17 +437,157 @@ getdnsify_addressdict(PyObject *pydict)
         PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
         return NULL;
     }
+#if PY_MAJOR_VERSION >= 3
+    if (!PyUnicode_Check(str))  {
+#else
     if (!PyString_Check(str))  {
+#endif
         PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
         return NULL;
     }
+#if PY_MAJOR_VERSION >= 3
+    if (inet_pton(domain, PyBytes_AsString(PyUnicode_AsEncodedString(str, "ascii", NULL)), buf) <= 0)  {
+#else
     if (inet_pton(domain, PyString_AsString(str), buf) <= 0)  {
+#endif
         PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
         return NULL;
     }
     addr_data.data = (uint8_t *)buf;
     addr_data.size = (domain == AF_INET ? 4 : 16);
     getdns_dict_set_bindata(addr_dict, "address_data", &addr_data);
+
+    if ((str = PyDict_GetItemString(pydict, "tls_auth_name")) != NULL)  {
+#if PY_MAJOR_VERSION >= 3
+        tls_auth_name.data = (uint8_t *)strdup(PyBytes_AsString(PyUnicode_AsEncodedString(str, "ascii", NULL)));
+#else
+        tls_auth_name.data = (uint8_t *)strdup(PyString_AsString(str));
+#endif
+        tls_auth_name.size = (size_t)strlen((char *)tls_auth_name.data);
+        getdns_dict_set_bindata(addr_dict, "tls_auth_name", &tls_auth_name);
+    }
+
+    if ((str = PyDict_GetItemString(pydict, "scope_id")) != NULL)  {
+#if PY_MAJOR_VERSION >= 3
+        scope_id.data = (uint8_t *)strdup(PyBytes_AsString(PyUnicode_AsEncodedString(str, "ascii", NULL)));
+#else
+        scope_id.data = (uint8_t *)strdup(PyString_AsString(str));
+#endif
+        scope_id.size = (size_t)strlen((char *)scope_id.data);
+        getdns_dict_set_bindata(addr_dict, "scope_id", &scope_id);
+    }
+
+    if ((str = PyDict_GetItemString(pydict, "tsig_name")) != NULL)  {
+#if PY_MAJOR_VERSION >= 3
+        tsig_name.data = (uint8_t *)strdup(PyBytes_AsString(PyUnicode_AsEncodedString(str, "ascii", NULL)));
+#else
+        tsig_name.data = (uint8_t *)strdup(PyBytes_AsString(str));
+#endif
+        tsig_name.size = (size_t)strlen((char *)tsig_name.data);
+        if ((ret = getdns_dict_set_bindata(addr_dict, "tsig_name", &tsig_name)) != GETDNS_RETURN_GOOD)  {
+            PyErr_SetString(getdns_error, "bad tsig name");
+            return NULL;
+        }
+    }
+
+    if ((str = PyDict_GetItemString(pydict, "tsig_algorithm")) != NULL)  {
+#if PY_MAJOR_VERSION >= 3
+        tsig_alg.data = (uint8_t *)strdup(PyBytes_AsString(PyUnicode_AsEncodedString(str, "ascii", NULL)));
+#else
+        tsig_alg.data = (uint8_t *)strdup(PyBytes_AsString(str));
+#endif
+        tsig_alg.size = (size_t)strlen((char *)tsig_alg.data);
+        if ((ret = getdns_dict_set_bindata(addr_dict, "tsig_algorithm", &tsig_alg)) != GETDNS_RETURN_GOOD)  {
+            PyErr_SetString(getdns_error, "bad tsig algorithm");
+            return NULL;
+        }
+    }
+
+    if ((str = PyDict_GetItemString(pydict, "tsig_secret")) != NULL)  {
+        tsig_secret.size = PyByteArray_Size(str);
+#if PY_MAJOR_VERSION >= 3
+        tsig_secret.data = (uint8_t *)strdup(PyBytes_AS_STRING(str));
+#else
+        tsig_secret.data = (uint8_t *)strdup(PyBytes_AsString(str));
+#endif
+        if ((ret = getdns_dict_set_bindata(addr_dict, "tsig_secret", &tsig_secret)) != GETDNS_RETURN_GOOD)  {
+            PyErr_SetString(getdns_error, "bad tsig secret");
+            return NULL;
+        }
+    }
+
+    if ((str = PyDict_GetItemString(pydict, "tls_pubkey_pinset")) != NULL)  {
+        Py_ssize_t pinset_len;
+        int  i;
+        PyObject *py_item;
+        char *str_item;
+        getdns_dict *pubkey_pin = 0;
+
+        if (!PyList_Check(str))  {
+            PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
+            return NULL;
+        }
+        pinset_len = PyList_Size(str);
+        if (pinset_len == 0)    {
+            PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
+            return NULL;
+        }
+        tls_pubkey_pinset = getdns_list_create();
+        for ( i = 0 ; i < (int)pinset_len ; i++ )  {
+            py_item = PyList_GetItem(str, (Py_ssize_t)i);
+#if PY_MAJOR_VERSION >= 3
+            str_item = PyBytes_AsString(PyUnicode_AsEncodedString(py_item, "ascii", NULL));
+#else
+            str_item = PyString_AsString(py_item);
+#endif
+            pubkey_pin = getdns_pubkey_pin_create_from_string(0, str_item);
+            if (pubkey_pin == NULL)  {
+                PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
+                return NULL;
+            }
+            if ((ret = getdns_list_set_dict(tls_pubkey_pinset, i, pubkey_pin)) != GETDNS_RETURN_GOOD)  {
+                PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
+                return NULL;
+            }
+        }
+        if (pubkey_pin)
+            getdns_dict_destroy(pubkey_pin);
+        getdns_dict_set_list(addr_dict, "tls_pubkey_pinset", tls_pubkey_pinset);
+    }
+
+    if ((str = PyDict_GetItemString(pydict, "port")) != NULL)  {
+#if PY_MAJOR_VERSION >= 3
+        if (!PyLong_Check(str))  {
+#else
+        if (!PyInt_Check(str))  {
+#endif
+            PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
+            return NULL;
+        }
+#if PY_MAJOR_VERSION >= 3
+        port = (uint32_t)PyLong_AsLong(str);
+#else
+        port = (uint32_t)PyInt_AsLong(str);
+#endif
+        getdns_dict_set_int(addr_dict, "port", port);
+    }
+
+    if ((str = PyDict_GetItemString(pydict, "tls_port")) != NULL)  {
+#if PY_MAJOR_VERSION >= 3
+        if (!PyLong_Check(str))  {
+#else
+        if (!PyInt_Check(str))  {
+#endif
+            PyErr_SetString(getdns_error, GETDNS_RETURN_INVALID_PARAMETER_TEXT);
+            return NULL;
+        }
+#if PY_MAJOR_VERSION >= 3
+        tls_port = (uint32_t)PyLong_AsLong(str);
+#else
+        tls_port = (uint32_t)PyInt_AsLong(str);
+#endif
+        getdns_dict_set_int(addr_dict, "tls_port", tls_port);
+    }
     return addr_dict;
 }
 
@@ -391,10 +645,19 @@ pythonify_address_list(getdns_list *list)
         }
         py_item = PyDict_New();
         PyDict_SetItemString(py_item, "address_data",
+#if PY_MAJOR_VERSION >= 3
+                             PyUnicode_FromString(inet_ntop(domain, (void *)a_address_data->data,
+                                                           (char *)paddr_buf, 256)));
+#else
                              PyString_FromString(inet_ntop(domain, (void *)a_address_data->data,
                                                            (char *)paddr_buf, 256)));
+#endif
         PyDict_SetItemString(py_item, "address_type",
+#if PY_MAJOR_VERSION >= 3
+                             PyUnicode_FromString((domain == AF_INET ? "IPv4" : "IPv6")));
+#else
                              PyString_FromString((domain == AF_INET ? "IPv4" : "IPv6")));
+#endif
         PyList_Append(py_list, py_item);
     }
     return py_list;
@@ -465,7 +728,11 @@ glist_to_plist(struct getdns_list *list)
                 PyErr_SetString(getdns_error, getdns_get_errorstr_by_id(ret));
                 return NULL;
             }
+#if PY_MAJOR_VERSION >= 3
+            py_int = PyLong_FromLong((long)localint);
+#else
             py_int = PyInt_FromLong((long)localint);
+#endif
             if (PyList_Append(py_list, py_int) == -1)  {
                 PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
                 return NULL;
@@ -494,6 +761,7 @@ glist_to_plist(struct getdns_list *list)
 }
 
 
+
 PyObject *
 gdict_to_pdict(struct getdns_dict *dict)
 {
@@ -517,6 +785,9 @@ gdict_to_pdict(struct getdns_dict *dict)
         PyErr_SetString(getdns_error, getdns_get_errorstr_by_id(ret));
         return NULL;
     }
+#if 0
+    printf("%s\n", getdns_pretty_print_dict(dict)); /* XXX */
+#endif
     py_dict = PyDict_New();
     (void)getdns_list_get_length(keys, &n_keys);
     for (i = 0 ; i < (int)n_keys ; i++ )  {
@@ -572,7 +843,11 @@ gdict_to_pdict(struct getdns_dict *dict)
                 PyErr_SetString(getdns_error, getdns_get_errorstr_by_id(ret));
                 return NULL;
             }
+#if PY_MAJOR_VERSION >= 3
+            py_localint = PyLong_FromLong((long)int_item);
+#else
             py_localint = PyInt_FromLong((long)int_item);
+#endif
             if (PyDict_SetItemString(py_dict, (char *)key_name->data, py_localint) == -1)  {
                 PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
                 return NULL;
@@ -585,7 +860,10 @@ gdict_to_pdict(struct getdns_dict *dict)
                 PyErr_SetString(getdns_error, getdns_get_errorstr_by_id(ret));
                 return NULL;
             }
-            if ((py_localbindata = convertBinData(bindata_item, "")) == 0)  {
+            if ((bindata_item == 0) || (bindata_item->data == 0) || bindata_item->size == 0)  {
+                break;
+            }
+            if ((py_localbindata = convertBinData(bindata_item, (char *)key_name->data)) == 0)  {
                 return NULL;
             }
             if (PyDict_SetItemString(py_dict, (char *)key_name->data, py_localbindata) == -1)  {
@@ -625,6 +903,7 @@ void error_exit(char* msg, getdns_return_t ret)
 // replies_tree
 
 // Taken from getdns source to do label checking
+
 static int
 priv_getdns_bindata_is_dname(struct getdns_bindata *bindata)
 {
@@ -652,7 +931,11 @@ convertBinData(getdns_bindata* data,
     if (data->size == 1 && data->data[0] == 0) {
         PyObject *a_string;
 
+#if PY_MAJOR_VERSION >= 3
+        if ((a_string = PyUnicode_FromString(".")) == NULL)  {
+#else
         if ((a_string = PyString_FromString(".")) == NULL)  {
+#endif
             PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
             return NULL;
         }
@@ -674,7 +957,11 @@ convertBinData(getdns_bindata* data,
     if (printable == 1) {
         PyObject *a_string;
 
-        if ((a_string = PyString_FromString((char *)data->data)) == NULL)  {
+#if PY_MAJOR_VERSION >= 3
+        if ((a_string = PyUnicode_FromStringAndSize((char *)data->data, (Py_ssize_t)(data->size-1))) == NULL)  {
+#else
+        if ((a_string = PyString_FromStringAndSize((char *)data->data, (Py_ssize_t)(data->size-1))) == NULL)  {
+#endif
             PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
             return NULL;
         }
@@ -689,7 +976,11 @@ convertBinData(getdns_bindata* data,
 
         if (getdns_convert_dns_name_to_fqdn(data, &dname)
             == GETDNS_RETURN_GOOD) {
+#if PY_MAJOR_VERSION >= 3
+            if ((dname_string = PyUnicode_FromString(dname)) != NULL)  {
+#else
             if ((dname_string = PyString_FromString(dname)) != NULL)  {
+#endif
                 return(dname_string);
             }  else  {
                 PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
@@ -708,7 +999,11 @@ convertBinData(getdns_bindata* data,
         char* ipStr = getdns_display_ip_address(data);
         if (ipStr) {
             PyObject *addr_string;
+#if PY_MAJOR_VERSION >= 3
+            if ((addr_string = PyUnicode_FromString(ipStr)) == NULL)  {
+#else
             if ((addr_string = PyString_FromString(ipStr)) == NULL)  {
+#endif
                 PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
                 return NULL;
             }
@@ -718,7 +1013,11 @@ convertBinData(getdns_bindata* data,
         uint8_t *blob = (uint8_t *)malloc(data->size);
         
         memcpy(blob, data->data, data->size);
+#if PY_MAJOR_VERSION >= 3                       
+        return (PyMemoryView_FromMemory((char *)blob, (Py_ssize_t)data->size, PyBUF_READ));
+#else
         return (PyBuffer_FromMemory(blob, (Py_ssize_t)data->size));
+#endif
     }
     return NULL;                /* should never get here .. */
 }
@@ -729,11 +1028,12 @@ PyObject *convertToList(struct getdns_list* list);
 PyObject*
 convertToDict(struct getdns_dict* dict) {
 
+    PyObject *resultsdict1;
+    
     if (!dict) {
         return 0;
     }
 
-    PyObject* resultsdict1;
     if ((resultsdict1 = PyDict_New()) == NULL)  {
         error_exit("Unable to allocate response dict", 0);
         return NULL;
@@ -754,13 +1054,20 @@ convertToDict(struct getdns_dict* dict) {
         getdns_list_get_bindata(names, i, &nameBin);
         getdns_data_type type;
         getdns_dict_get_data_type(dict, (char*)nameBin->data, &type);
+
         switch (type) {
             case t_bindata:
             {
                 getdns_bindata* data = NULL;
                 getdns_dict_get_bindata(dict, (char*)nameBin->data, &data);
                 PyObject* res = convertBinData(data, (char*)nameBin->data);
-                PyDict_SetItem(resultsdict1, PyString_FromString((char*) nameBin->data), res);
+#if PY_MAJOR_VERSION >= 3
+                PyDict_SetItem(resultsdict1, PyUnicode_FromStringAndSize((char *)nameBin->data, 
+                                                                         (Py_ssize_t)nameBin->size), res);
+#else
+                PyDict_SetItem(resultsdict1, PyString_FromStringAndSize((char *)nameBin->data,
+                                                                 (Py_ssize_t)nameBin->size), res);
+#endif
                 break;
             }
             case t_int:
@@ -769,7 +1076,13 @@ convertToDict(struct getdns_dict* dict) {
                 getdns_dict_get_int(dict, (char*)nameBin->data, &res);
                 PyObject* rl1 = Py_BuildValue("i", res);
                 PyObject *res1 = Py_BuildValue("O", rl1);
-                PyDict_SetItem(resultsdict1, PyString_FromString((char*) nameBin->data), res1);
+#if PY_MAJOR_VERSION >= 3
+                PyDict_SetItem(resultsdict1, PyUnicode_FromStringAndSize((char *)nameBin->data,
+                                                                         (Py_ssize_t)nameBin->size), res1);
+#else
+                PyDict_SetItem(resultsdict1, PyString_FromStringAndSize((char *)nameBin->data,
+                                                                        (Py_ssize_t)nameBin->size), res1);
+#endif
                 break;
             }
             case t_dict:
@@ -778,7 +1091,13 @@ convertToDict(struct getdns_dict* dict) {
                 getdns_dict_get_dict(dict, (char*)nameBin->data, &subdict);
                 PyObject *rl1 = convertToDict(subdict);
                 PyObject *res1 = Py_BuildValue("O", rl1);
-                PyDict_SetItem(resultsdict1, PyString_FromString((char*) nameBin->data), res1);
+#if PY_MAJOR_VERSION >= 3
+                PyDict_SetItem(resultsdict1, PyUnicode_FromStringAndSize((char *)nameBin->data,
+                                                                         (Py_ssize_t)nameBin->size), res1);
+#else
+                PyDict_SetItem(resultsdict1, PyString_FromStringAndSize((char *)nameBin->data,
+                                                                        (Py_ssize_t)nameBin->size), res1);
+#endif
                 break;
             }
             case t_list:
@@ -787,7 +1106,13 @@ convertToDict(struct getdns_dict* dict) {
                 getdns_dict_get_list(dict, (char*)nameBin->data, &list);
                 PyObject *rl1 = convertToList(list);
                 PyObject *res1 = Py_BuildValue("O", rl1);
-                PyObject *key = PyString_FromString((char *)nameBin->data);
+#if PY_MAJOR_VERSION >= 3
+                PyObject *key = PyUnicode_FromStringAndSize((char *)nameBin->data,
+                                                            (Py_ssize_t)nameBin->size);
+#else
+                PyObject *key = PyString_FromStringAndSize((char *)nameBin->data,
+                                                           (Py_ssize_t)nameBin->size);
+#endif
                 PyDict_SetItem(resultsdict1, key, res1);
                 break;
             }
@@ -898,12 +1223,22 @@ getdns_dict_to_ip_string(getdns_dict* dict) {
             PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
             return NULL;
         }
-        if (PyDict_SetItemString(addr_dict, "address_type", PyString_FromString((char *)type->data)) != 0)  {
+#if PY_MAJOR_VERSION >= 3
+        if (PyDict_SetItemString(addr_dict, "address_type", PyUnicode_FromStringAndSize((char *)type->data,
+                                                                                        (Py_ssize_t)type->size)) != 0)  {
+#else
+        if (PyDict_SetItemString(addr_dict, "address_type", PyString_FromStringAndSize((char *)type->data,
+                                                                                       (Py_ssize_t)type->size)) != 0)  {
+#endif
             PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
             return NULL;
         }
         addr_string = getdns_display_ip_address(addr);
+#if PY_MAJOR_VERSION >= 3
+        if ((pyaddr_string = PyUnicode_FromString(addr_string)) == NULL)  {
+#else
         if ((pyaddr_string = PyString_FromString(addr_string)) == NULL)  {
+#endif
             PyErr_SetString(getdns_error, GETDNS_RETURN_GENERIC_ERROR_TEXT);
             return NULL;
         }
@@ -915,71 +1250,4 @@ getdns_dict_to_ip_string(getdns_dict* dict) {
     return NULL;
 }
 
-PyObject*
-getFullResponse(struct getdns_dict *dict)
-{
-
-    PyObject* resultslist;
-    if ((resultslist = PyDict_New()) == NULL)  {
-        error_exit("Unable to allocate response list", 0);
-        return NULL;
-    }
-
-    getdns_list* names;
-    getdns_dict_get_names(dict, &names);
-    size_t len = 0, i = 0;
-    getdns_list_get_length(names, &len);
-    for (i = 0; i < len; ++i) {
-        getdns_bindata* nameBin;
-        getdns_list_get_bindata(names, i, &nameBin);
-        getdns_data_type type;
-        getdns_dict_get_data_type(dict, (char*)nameBin->data, &type);
-        switch (type) {
-            case t_bindata:
-            {
-                getdns_bindata* data = NULL;
-                getdns_dict_get_bindata(dict, (char*)nameBin->data, &data);
-                PyObject *res = convertBinData(data, (char*)nameBin->data);
-                if (res) {
-                   PyDict_SetItem(resultslist, PyString_FromString((char*)nameBin->data), res);
-                } else {
-                   PyObject* res1 = Py_BuildValue("s", "empty");
-                   PyDict_SetItem(resultslist, PyString_FromString((char*)nameBin->data), res1);
-                }
-                break;
-            }
-            case t_int:
-            {
-                uint32_t res = 0;
-                getdns_dict_get_int(dict, (char*)nameBin->data, &res);
-                PyObject* res1 = Py_BuildValue("i", res);
-                PyDict_SetItem(resultslist, PyString_FromString((char*)nameBin->data), res1);
-                break;
-            }
-            case t_dict:
-            {
-                getdns_dict* subdict = NULL;
-                getdns_dict_get_dict(dict, (char*)nameBin->data, &subdict);
-                PyObject* rl1 = convertToDict(subdict);
-                PyObject *res1 = Py_BuildValue("O", rl1);
-                PyDict_SetItem(resultslist, PyString_FromString((char*)nameBin->data), res1);
-                break;
-            }
-            case t_list:
-            {
-                getdns_list* list = NULL;
-                getdns_dict_get_list(dict, (char*)nameBin->data, &list);
-                PyObject* rl1 = convertToList(list);
-                PyDict_SetItem(resultslist, PyString_FromString((char*)nameBin->data), rl1); 
-                break;
-            }
-            default:
-                break;
-        }
-    }
-
-    getdns_list_destroy(names);
-
-    return resultslist;
-}
 
